@@ -8,6 +8,7 @@ from glob import has_magic
 import logging
 import os
 import warnings
+import weakref
 
 from azure.core.exceptions import ResourceNotFoundError, ResourceExistsError
 from azure.storage.blob._shared.base_client import create_configuration
@@ -25,10 +26,26 @@ from fsspec.asyn import (
     sync_wrapper,
 )
 from fsspec.utils import infer_storage_options, tokenize
+from .utils import filter_blobs
 
 
 logger = logging.getLogger(__name__)
 
+FORWARDED_BLOB_PROPERTIES = [
+    "metadata",
+    "creation_time",
+    "deleted",
+    "deleted_time",
+    "last_modified",
+    "content_time",
+    "content_settings",
+    "remaining_retention_days",
+    "archive_status",
+    "last_accessed_on",
+    "tags",
+    "tag_count",
+]
+
 
 class AzureDatalakeFileSystem(AbstractFileSystem):
     """
@@ -381,6 +398,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         else:
             self.sync_credential = None
         self.do_connect()
+        weakref.finalize(self, maybe_sync, self.service_client.close, self)
 
     @classmethod
     def _strip_protocol(cls, path: str):
@@ -397,7 +415,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         str
             Returns a path without the protocol
         """
-        logging.debug(f"_strip_protocol for {path}")
+        logger.debug(f"_strip_protocol for {path}")
         ops = infer_storage_options(path)
 
         # we need to make sure that the path retains
@@ -407,7 +425,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             ops["path"] = ops["host"] + ops["path"]
         ops["path"] = ops["path"].lstrip("/")
 
-        logging.debug(f"_strip_protocol({path}) = {ops}")
+        logger.debug(f"_strip_protocol({path}) = {ops}")
         return ops["path"]
 
     def _get_credential_from_service_principal(self):
@@ -666,7 +684,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         return_glob: bool
 
         """
-        logging.debug(f"abfs.ls() is searching for {path}")
+        logger.debug(f"abfs.ls() is searching for {path}")
         target_path = path.strip("/")
         container, path = self.split_path(path)
 
@@ -676,7 +694,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         if (container in ["", ".", delimiter]) and (path in ["", delimiter]):
             if path not in self.dircache or invalidate_cache or return_glob:
                 # This is the case where only the containers are being returned
-                logging.info(
+                logger.info(
                     "Returning a list of containers in the azure blob storage account"
                 )
                 contents = self.service_client.list_containers(include_metadata=True)
@@ -693,7 +711,9 @@ class AzureBlobFileSystem(AsyncFileSystem):
                         container=container
                     )
                     path = path.strip("/")
-                    blobs = container_client.walk_blobs(name_starts_with=path)
+                    blobs = container_client.walk_blobs(
+                        include=["metadata"], name_starts_with=path
+                    )
 
                     # Check the depth that needs to be screened
                     depth = target_path.count("/")
@@ -727,10 +747,14 @@ class AzureBlobFileSystem(AsyncFileSystem):
                     except ResourceNotFoundError:
                         raise FileNotFoundError
                     if return_glob:
-                        finalblobs = await self._details(outblobs, return_glob=True)
+                        finalblobs = await self._details(
+                            outblobs, target_path=target_path, return_glob=True
+                        )
                         return finalblobs
                     else:
-                        finalblobs = await self._details(outblobs)
+                        finalblobs = await self._details(
+                            outblobs, target_path=target_path
+                        )
                     if not finalblobs:
                         if not await self._exists(target_path):
                             raise FileNotFoundError
@@ -741,7 +765,12 @@ class AzureBlobFileSystem(AsyncFileSystem):
             return self.dircache[target_path]
 
     async def _details(
-        self, contents, delimiter="/", return_glob: bool = False, **kwargs
+        self,
+        contents,
+        delimiter="/",
+        return_glob: bool = False,
+        target_path="",
+        **kwargs,
     ):
         """
         Return a list of dictionaries of specifying details about the contents
@@ -763,7 +792,11 @@ class AzureBlobFileSystem(AsyncFileSystem):
 
         output = []
         for content in contents:
-            data = {}
+            data = {
+                key: content[key]
+                for key in FORWARDED_BLOB_PROPERTIES
+                if content.has_key(key)  # NOQA
+            }
             if content.has_key("container"):  # NOQA
                 fname = f"{content.container}{delimiter}{content.name}"
                 data.update({"name": fname})
@@ -783,10 +816,17 @@ class AzureBlobFileSystem(AsyncFileSystem):
             if return_glob:
                 data.update({"name": data["name"].rstrip("/")})
             output.append(data)
+        if target_path:
+            if len(output) == 1 and output[0]["type"] == "file":
+                # This handles the case where path is a file passed to ls
+                return output
+            output = await filter_blobs(output, target_path)
         return output
 
     def find(self, path, maxdepth=None, withdirs=False, **kwargs):
-        return maybe_sync(self._find, self, path, maxdepth, withdirs)
+        return maybe_sync(
+            self._find, self, path=path, maxdepth=maxdepth, withdirs=withdirs, **kwargs
+        )
 
     async def _find(self, path, maxdepth=None, withdirs=False, **kwargs):
         """List all files below path.
@@ -858,8 +898,6 @@ class AzureBlobFileSystem(AsyncFileSystem):
             listing = await self._ls(path, return_glob=True, **kwargs)
         except (FileNotFoundError, IOError):
             listing = []
-            # self._stop_iterating()
-            # yield [], [], []
 
         for info in listing:
             # each info name must be at least [path]/part , but here
@@ -894,7 +932,14 @@ class AzureBlobFileSystem(AsyncFileSystem):
                 yield path, dirs, files
 
     def mkdir(self, path, delimiter="/", exist_ok=False, **kwargs):
-        maybe_sync(self._mkdir, self, path, delimiter, exist_ok)
+        maybe_sync(
+            self._mkdir,
+            self,
+            path=path,
+            delimiter=delimiter,
+            exist_ok=exist_ok,
+            **kwargs,
+        )
 
     async def _mkdir(self, path, delimiter="/", exist_ok=False, **kwargs):
         """
@@ -993,7 +1038,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
                 container_client = self.service_client.get_container_client(
                     container=container_name
                 )
-                logging.debug(f"Delete blob {path} in {container_name}")
+                logger.debug(f"Delete blob {path} in {container_name}")
                 await container_client.delete_blob(path)
             elif kind == "directory":
                 container_name, path = self.split_path(path, delimiter=delimiter)
@@ -1003,7 +1048,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
                 _containers = await self._ls("")
                 _containers = [c["name"] for c in _containers]
                 if (container_name + delimiter in _containers) and (not path):
-                    logging.debug(f"Delete container {container_name}")
+                    logger.debug(f"Delete container {container_name}")
                     await container_client.delete_container()
             else:
                 raise RuntimeError(f"Unable to delete {path}!")
@@ -1237,6 +1282,22 @@ class AzureBlobFileSystem(AsyncFileSystem):
 
     get_file = sync_wrapper(_get_file)
 
+    def getxattr(self, path, attr):
+        meta = self.info(path).get("metadata", {})
+        return meta[attr]
+
+    async def _setxattrs(self, rpath, **kwargs):
+        container_name, path = self.split_path(rpath)
+        try:
+            cc = self.service_client.get_container_client(container_name)
+            bc = cc.get_blob_client(blob=path)
+            await bc.set_blob_metadata(metadata=kwargs)
+            self.invalidate_cache(self._parent(rpath))
+        except Exception as e:
+            raise FileNotFoundError(f"File not found for {e}")
+
+    setxattrs = sync_wrapper(_setxattrs)
+
     def invalidate_cache(self, path=None):
         if path is None:
             self.dircache.clear()
@@ -1252,6 +1313,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
         autocommit: bool = True,
         cache_options: dict = {},
         cache_type="readahead",
+        metadata=None,
         **kwargs,
     ):
         """Open a file on the datalake, or a block blob
@@ -1276,7 +1338,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             See the definitions here:
             https://filesystem-spec.readthedocs.io/en/latest/api.html#readbuffering
         """
-        logging.debug(f"_open:  {path}")
+        logger.debug(f"_open:  {path}")
         return AzureBlobFile(
             fs=self,
             path=path,
@@ -1285,6 +1347,7 @@ class AzureBlobFileSystem(AsyncFileSystem):
             autocommit=autocommit,
             cache_options=cache_options,
             cache_type=cache_type,
+            metadata=metadata,
             **kwargs,
         )
 
@@ -1303,6 +1366,7 @@ class AzureBlobFile(io.IOBase):
         autocommit: bool = True,
         cache_type: str = "readahead",
         cache_options: dict = {},
+        metadata=None,
         **kwargs,
     ):
         """
@@ -1369,7 +1433,7 @@ class AzureBlobFile(io.IOBase):
                 FutureWarning,
             )
             cache_options["trim"] = kwargs.pop("trim")
-
+        self.metadata = None
         self.kwargs = kwargs
         self.connect_client()
         if self.mode not in {"ab", "rb", "wb"}:
@@ -1382,6 +1446,7 @@ class AzureBlobFile(io.IOBase):
                 self.blocksize, self._fetch_range, self.size, **cache_options
             )
         else:
+            self.metadata = metadata
             self.buffer = io.BytesIO()
             self.offset = None
             self.forced = False
@@ -1511,7 +1576,7 @@ class AzureBlobFile(io.IOBase):
         elif self.mode == "ab":
             self.blob_client = self.container_client.get_blob_client(blob=self.blob)
             if not self.fs.exists(self.path):
-                self.blob_client.create_append_blob()
+                self.blob_client.create_append_blob(metadata=self.metadata)
         else:
             raise ValueError(
                 "File operation modes other than wb are not yet supported for writing"
@@ -1535,24 +1600,29 @@ class AzureBlobFile(io.IOBase):
         if self.mode == "wb":
             try:
                 self.blob_client.stage_block(
-                    block_id=block_id, data=data, length=length
+                    block_id=block_id, data=data, length=length,
                 )
                 self._block_list.append(block_id)
 
                 if final:
                     block_list = [BlobBlock(_id) for _id in self._block_list]
-                    self.blob_client.commit_block_list(block_list=block_list)
+                    self.blob_client.commit_block_list(
+                        block_list=block_list, metadata=self.metadata
+                    )
             except Exception as e:
                 # This step handles the situation where data="" and length=0
                 # which is throws an InvalidHeader error from Azure, so instead
                 # of staging a block, we directly upload the empty blob
                 if block_id == "0000000" and length == 0 and final:
-                    self.blob_client.upload_blob(data=data)
+                    self.blob_client.upload_blob(data=data, metadata=self.metadata)
                 else:
                     raise RuntimeError(f"Failed to upload block with {e}!!")
         elif self.mode == "ab":
             self.blob_client.upload_blob(
-                data=data, length=length, blob_type=BlobType.AppendBlob
+                data=data,
+                length=length,
+                blob_type=BlobType.AppendBlob,
+                metadata=self.metadata,
             )
 
     def flush(self, force=False):
